/* 
 *
 * Dumps process vmspace (except vm_pages) in human
 * readable format.
 *
 * $Id: dump_vmspace.c,v 1.3 2006/06/22 11:55:17 romanp Exp $
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/syscall.h>
#include <sys/module.h>
#include <sys/user.h>
#include <sys/vnode.h>

#include <vm/vm_object.h>
#include <vm/swap_pager.h>

#include <kvm.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sysexits.h>
#include <string.h>

#include "kernel.h"

u_int m_size, o_size, r_size; /* total process maps, objs, resident pages */
u_int p_size_m, p_size_r;     /* private maps, resident pages */
u_int s_size_m, s_size_r;     /* shared maps, resident pages */
u_int swap_size;              /* process swap usage */

/* total statistics for all processes */
u_int total_swap, total_rsize, total_osize, total_msize, total_psizer;

typedef struct _bit_t {
    char 	name[32];
    u_long 	mask;
} bit_t;

/* meanings of vm_map_entry_t->eflags */
bit_t vme_flags[] = {
		{"NS",		0x0001},
		{"SMAP", 	0x0002},
		{"COW",		0x0004},
		{"NC",		0x0008},
		{"NFLT",	0x0010},
		{"UW",		0x0020},
		{"BE_SEQ",	0x0040},
		{"BE_RND",	0x0080},
		{"IN_TRANS",0x0100},
		{"NWAKEUP",	0x0200},
		{"NOCORE",	0x0400},
		{"GDOWN",	0x1000},
		{"GUP",		0x2000},
		{"BE_NOR",	0x0000}
};

/* meanings of vm_object_t->flags */
bit_t	vmobj_flags[] = { 
    {"ACT",0x0004},
    {"DEAD",0x0008},
    {"NSPLIT",0x0010},
    {"PIPWNT",0x0040},
    {"WRITBLE",0x0080},
    {"MDIRTY",0x0100},
    {"CLEANING",0x0200},
    {"ONEMAP",0x2000},
    {"",0},
};

bit_t	vmmap_flags[] = { 
    {"WIREFUTURE",0x0001},
};

int 		main(int argc, char **argv);
static      void process_vmobject(u_long objaddr);
static      char *make_vmo_type(u_char);
static      char *print_bits(uint32_t, bit_t *);

int 
main(int argc, char **argv)
{
	struct 	vmspace	vmspace;
	struct	vm_map	*vmm;
	struct	vm_map_entry vme;
	struct	kinfo_proc *kip;
	int		i;

    if(argc != 2) {
		/* use argv[1] as pid */
		printf("usage: dump_vmspace pid\n");
		exit(EX_USAGE);
	}

    /* open kernel memory */
    if(k_open() == -1) 
		exit(1);

    /* get process entry */
    kip = k_getproc(atoi(argv[1]));
    if(kip == NULL)
		exit(EX_OSERR);
	/* print general info */
	printf("\npid: %i comm: %s\n", kip->ki_pid, kip->ki_comm);
	printf("VM Space located at %p\n", kip->ki_vmspace);
	k_read((u_long)kip->ki_vmspace, &vmspace, sizeof(vmspace));
	vmm = &vmspace.vm_map;
	printf("VM Map: version: %d, size: %d, %d entries, system_map: %d, flags: %s\n",
		vmm->timestamp, vmm->size, vmm->nentries,
		vmm->system_map, print_bits(vmm->flags, vmmap_flags));

	/* walk through vm_map_entry chain */
	printf("end start object offset wired prot flags\n");
	for(vme = vmm->header, i = 0; i < vmm->nentries; i++) {
		k_read((u_long)vme.next, &vme, sizeof(vme));
		printf("%d -> 0x%lx - 0x%lx 0x%lx %lx %d %s%s%s %s\n",
			i, (u_long)vme.start, (u_long)vme.end,
			(u_long)vme.object.vm_object, (u_long)vme.offset,
			vme.wired_count,
			(vme.protection & VM_PROT_READ)?"r":"-",
			(vme.protection & VM_PROT_WRITE)?"w":"-",
			(vme.protection & VM_PROT_EXECUTE)?"x":"-",
			print_bits(vme.eflags, vme_flags));
		if (vme.object.vm_object != NULL)
			process_vmobject((u_long)vme.object.vm_object);
	}

    return (0);
}

static void
process_vmobject(u_long objaddr)
{
	struct vm_object	vmo;

	printf("\taddr type size ref shadow res bobj flags\n");

	do {
		k_read(objaddr, &vmo, sizeof(vmo));
		printf("\t0x%lx: %s %lu %d %d %d %p %s\n",
			objaddr, make_vmo_type(vmo.type), (u_long)vmo.size,
			vmo.ref_count, vmo.shadow_count, vmo.resident_page_count,
			vmo.backing_object, print_bits(vmo.flags, vmobj_flags));
		objaddr = (u_long)vmo.backing_object;
	} while(objaddr != 0);
}

static char *
print_bits(uint32_t reg, bit_t *bits)
{
    int 	i, first = 1;
    static 	char buff[1024];

    sprintf(buff,"0x%x<", reg);
    for(i=0;;i++) {
		if(!bits[i].name[0])
			break;
		if(reg & bits[i].mask) {
			if(!first)
				strcat(buff,",");
			else
				first = 0;
			strcat(buff,bits[i].name);
		}
    }
    strcat(buff,">");

    return buff;
}

static char *
make_vmo_type(objtype_t t)
{
    static char *types[]= {    "DEF", "SWP", "VN",
                               "DEV", "PHY", "DEAD" };
    return types[(int)t];
}
