/*-
 * Copyright (c) 2000, 2002, 2006 Roman V. Palagin
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__RCSID("$Id: pmap.c,v 1.11 2006/06/22 11:55:17 romanp Exp $");

#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/syscall.h>
#include <sys/module.h>
#include <sys/user.h>
#include <sys/vnode.h>
#include <sys/linker.h>

#include <vm/vm_object.h>
#include <vm/swap_pager.h>

#include <kvm.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <regex.h>
#include <stdint.h>

#include "pmap_helper.h"
#include "kernel.h"

/* total memory usage for process */
u_long p_msize = 0, p_rsize = 0, p_ssize = 0, p_psize = 0;

/* process memory map size */
u_long	map_size;

int 		main(int argc, char **argv);
static void 	process_map(struct pmh_map *);
static void	print_proto(vm_prot_t);
static void	usage(void);

/*
 * Format vars
 */
const char      *bar;
const char *bar8 = "-------";
const char *bar16 = "----------";
int     	a_wdth, s_wdth;

#define MAP_VME(m)  	((m).entry)
#define NKPG		(PAGE_SIZE/1024)

int 
main(int argc, char **argv)
{
	int c, nproc, i, prc;
	pid_t pid = 0;
	struct module_stat stat;
	int modid;
	char	*prog_wild = NULL, *prog_name;
	regex_t preg;
	struct pargs *pa = NULL;
    struct kinfo_proc *kp;
    int	pmap_helper_syscall;
	
    if ((modid = modfind("pmap_helper")) == -1) { 
		/* module not found, try to load */
		modid = kldload("pmap_helper.ko");
		if (modid == -1) 
			err(1, "unable to load pmap_helper module");
		modid = modfind("pmap_helper");
		if (modid == -1)
			err(1, "pmap_helper module loaded but not found");
	}

	stat.version = sizeof(stat);
	modstat(modid, &stat);
	pmap_helper_syscall = stat.data.intval;

	while ((c = getopt(argc, argv, "n:")) != EOF) {
		switch (c) {
		case 'n':
			prog_wild = optarg;
			break;
		default:
			usage();
			break;
		}
	}
	if (prog_wild != NULL) {
		if (0 != regcomp(&preg, prog_wild,
			REG_EXTENDED | REG_ICASE | REG_NOSUB)) {
			errx(1, "regcomp() failed\n");
		}
	} else if ((argc != 2) || ((pid = atoi(argv[1])) == 0)) {
		usage();
	}

	/* open kernel memory */
	k_open();

	/* read process list */
	kp = k_getallprocs(&nproc);

	for (prc = 0; prc < nproc; prc++) {
		struct pmh		pmh;
		struct pmh_map 	maps[1024];

		if ((pid != 0) && (kp[prc].ki_pid != pid))
			continue;
		/* try to read command line */
		if (kp[prc].ki_args != NULL) {
			int rc;

			pa = realloc(pa, sizeof(struct pargs));
			rc = k_read((u_long)kp[prc].ki_args, pa, sizeof(struct pargs));
			if (rc != -1) {
				int news = sizeof(struct pargs) + pa->ar_length + 1;
				pa = realloc(pa, news);
				memset(pa, 0, news);
				rc = k_read((u_long)kp[prc].ki_args, pa, news + 1);
			}
			prog_name = (char*)pa->ar_args;
		} else {
			prog_name = kp[prc].ki_comm;

		}

		if (prog_wild != NULL) {
			if (regexec(&preg, prog_name, 0, NULL, 0) == REG_NOMATCH)
				continue;
		}

		/* print general info */
		printf("%i:\t%s\n", kp[prc].ki_pid, prog_name);
		/* get virtual maps for process via pmap_helper */
		pmh.version = PMH_VERSION;
		pmh.pid = kp[prc].ki_pid;
		pmh.nmaps = 1024;
		pmh.maps = maps;
		if (syscall(pmap_helper_syscall, &pmh)) {
			warn("pmap_helper");
			continue;
		}
		/*
		 * Set format strings/width
		 */
		/*
		 * ki_emul contains so called ABI name for given
		 * process. If string contains 'ELF64' we assume
		 * LP64 mode, if not - ILP32.
		 */
		 if (strstr(kp[prc].ki_emul, "ELF64") != NULL) {
			/* process is 64bit ELF executable */
			a_wdth = 16;
			s_wdth = 10;
			bar = bar16;
		 } else {
			a_wdth = 8;
			s_wdth = 7;
			bar = bar8;
		 }

		printf("%-*s %*s %*s %*s %*s %-5s %s\n",
		    a_wdth, "Address", s_wdth, "Kbytes",
	            s_wdth, "RSS", s_wdth, "Shared",
		    s_wdth, "Priv", 
		    "Mode", "Mapped File");
		p_msize = p_rsize = p_ssize = p_psize = 0;
	
		for (i=0; i < pmh.nmaps; printf("\n"), i++) {
			/*
			 * Skips maps with no lower objects
			 */
			if (!maps[i].lobj_present)
				continue;
			map_size = (MAP_VME(maps[i]).end - MAP_VME(maps[i]).start)/1024;
			printf("%0*lX %*lu ", a_wdth, (u_long)MAP_VME(maps[i]).start, s_wdth, map_size);
			p_msize += map_size;
			process_map(&maps[i]);
		}
		printf("%s%s %s %s %s %s\n",
		    a_wdth == 8 ? "-" : "------", bar, bar, bar, bar, bar);
		printf("%-*s %*lu %*lu %*lu %*lu\n\n",
		    a_wdth, "Total Kb", s_wdth, p_msize, s_wdth, p_rsize,
		    s_wdth, p_ssize, s_wdth, p_psize);
	}
		
	k_close();
	if (prog_wild != NULL)
		regfree(&preg);
	exit(0);
}

static void 
process_map(struct pmh_map *map)
{
	struct vm_object *obj = &map->obj;
	struct vm_object *lobj = &map->lobj;
	struct pmh_map m = *map;
	u_long pkbytes;

	if (!m.obj_present)
		printf("oops - no obj!");

	/* 
	 * we calculate memory usage depending on object type
	 * output: resident, shared, priv, proto, mapped file 
	 */
	switch (lobj->type) {
	case OBJT_DEFAULT: /* XXX treat all memory as private. Is it 
			      always correct? 
			   */
		printf("%*lu %*s %*lu ", s_wdth, m.resident_pages * NKPG, 
		    s_wdth, "-", s_wdth, map_size); 
		p_rsize += m.resident_pages * NKPG;
		p_psize += map_size;
		print_proto(MAP_VME(m).protection);       
		printf("  [ anon ]");
		break;

	case OBJT_VNODE:
	        pkbytes = (obj->shadow_count == 1) ? 
		  (u_long)(obj->resident_page_count * NKPG) :
	           m.resident_pages * NKPG ;
		printf("%*lu ", s_wdth, pkbytes); 
		p_rsize +=  pkbytes;
		if (obj->shadow_count == 1) {
		  printf("%*s %*lu ", s_wdth, "-", s_wdth, pkbytes);
			p_psize += pkbytes;
		} else if (obj->ref_count == 1) { /* XXX does it alawys mean private mapping (?) */
			printf("%*s %*lu ", s_wdth, "-", s_wdth, map_size);
			p_psize += map_size;
		} else {
			printf("%*lu %*s ", s_wdth, map_size, s_wdth, "-");
			p_ssize += map_size;
		}
		print_proto(MAP_VME(m).protection);       
        printf("%s", m.fname);
		break;

	case OBJT_SWAP:
		printf("%*s %*s %*s ", s_wdth, "-", s_wdth, "-", s_wdth, "-");
		print_proto(MAP_VME(m).protection);       
		printf("[swap pager]"); 
		break;

	case OBJT_DEVICE:
		printf("%*lu %*s %*lu ", s_wdth, m.resident_pages * NKPG,
                       s_wdth, "-", s_wdth, map_size);
		print_proto(MAP_VME(m).protection);
		printf("/dev/%s", m.fname);
		break;

	case OBJT_DEAD:
                printf("%*s %*s %*s ", s_wdth, "-", s_wdth, "-", s_wdth, "-");
		print_proto(MAP_VME(m).protection);       
		printf("[dead object]");
		break;

	default:
		printf("[unknown (%i) object type]", lobj->type);
		break;
	}
}

static void
print_proto (vm_prot_t p)
{
	char *str;
	int i = 0;

	str = strdup("---");

	if(p & VM_PROT_READ)
		str[i] = 'r';
	i++;	
	if(p & VM_PROT_WRITE)
		str[i] = 'w';
	i++;
	if(p & VM_PROT_EXECUTE)
		str[i] = 'x';
	printf("%-5s ", str);
	free(str);
}

static void
usage(void)
{
	errx(1, "usage: pmap -n regexp | pid\n");
}
