/*-
 * Copyright (c) 2000, 2002, 2006 Roman V. Palagin
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Current version of this file is commented according to
 * "Design and Implementation of FreeBSD".
 *
 * $Id: pmap_helper.c,v 1.11 2006/06/22 11:55:17 romanp Exp $
 */

#include <sys/cdefs.h>
__RCSID("$Id: pmap_helper.c,v 1.11 2006/06/22 11:55:17 romanp Exp $");

#include <sys/types.h>
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/proc.h>
#include <sys/module.h>
#include <sys/sysent.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/lock.h>
#include <sys/vnode.h>
#include <sys/malloc.h>

#include <vm/vm.h>
#include <vm/pmap.h>
#include <vm/vm_map.h>
#include <vm/vm_page.h>
#include <vm/vm_object.h>
#include <vm/swap_pager.h>

#include "pmap_helper.h"

/* XXX From /sys/kern/makesyscalls.sh */
#define AS(type) (sizeof(struct type)/sizeof(register_t))

struct pmap_helper_args {
	struct pmh 		*pmh;
};

static int pmap_helper(struct proc *, struct pmap_helper_args *);
static int pmap_helper_modevt(struct module *, int, void *);

static struct sysent pmap_helper_sysent = {
    AS(pmap_helper_args),	/* sy_narg */
    (sy_call_t *)pmap_helper 	/* sy_call */
};

static int offset = NO_SYSCALL; /* offset in syscall table */

/* this is syscall himself */
static int
pmap_helper(struct proc *curp, struct pmap_helper_args *uap)
{
	struct proc *p;
	static struct vm_map *map;
	struct vm_map_entry *entry;
	pmap_t pmap;
	int nmaps = 0;
	int error;
	struct pmh pmh;
	struct pmh_map pmhm;
	size_t done;

	if ((error = copyin(uap->pmh, &pmh, sizeof pmh)))
		return error;

	if (pmh.version != PMH_VERSION)
		return EINVAL;

	p = pfind(pmh.pid); /* XXX Locks p! */
	if (p == NULL)
		return ESRCH;

	/* map points to description of MI virtual address space */
	map = &p->p_vmspace->vm_map;

       if (map != &curthread->td_proc->p_vmspace->vm_map)
               vm_map_lock_read(map);
 
	/* pmap points to private physiscal map for whole process space */
	pmap = vmspace_pmap(p->p_vmspace);
	error = 0;

	for (entry = map->header.next; entry != &map->header; 
	     entry = entry->next) {

		struct vm_object *obj, *tobj, *lobj;
		vm_offset_t addr;
		int resident;
	
		if (entry->eflags & MAP_ENTRY_IS_SUB_MAP)
			continue;

		bzero(&pmhm, sizeof pmhm);

		/* save entry */
		pmhm.entry = *entry;

		/* 
		 * obj points to chain of vm_object structures
		 * that describes sources of data that are mapped
		 * at the address range
		 */
		obj = entry->object.vm_object;
		resident = 0;
		addr = entry->start;
		/* estimate number of phys. resident pages in map */
		while (addr < entry->end) {
			if (pmap_extract(pmap, addr))
				resident++;
			addr += PAGE_SIZE;
		}

		pmhm.resident_pages = resident;

		/* move to lowest object which representing data source */
		for (lobj = tobj = obj; tobj; tobj = tobj->backing_object)
			lobj = tobj;
		if (obj) {
			pmhm.obj_present = 1;
			pmhm.obj = *obj;
		}
		if (lobj) {
			pmhm.lobj_present = 1;
			pmhm.lobj = *lobj;

			if (lobj->type == OBJT_DEVICE) {
			        struct cdevsw *csw;
	
				csw = dev_refthread(lobj->handle);
				copystr(csw->d_name, pmhm.fname, 254, &done);
				dev_relthread(lobj->handle);
			} else if (lobj->type == OBJT_VNODE) {
				char	*fullpath = "[vnode pager]", *freepath = NULL;

				vn_fullpath(curthread, (struct vnode*)lobj->handle,
					&fullpath, &freepath);
				copystr(fullpath, pmhm.fname, 1024, &done);
				if (freepath != NULL)
					free(freepath, M_TEMP);
			}
		}
		error = copyout(&pmhm, &pmh.maps[nmaps], sizeof pmh.maps[nmaps]);
		if (error)
			break;
		nmaps++;
		if (nmaps > pmh.nmaps)
			break;
	}

        if (map != &curthread->td_proc->p_vmspace->vm_map)
               vm_map_unlock_read(map);

	PROC_UNLOCK(p);	/* locked by pfind(9) */

	if (error == 0)
		error = copyout(&nmaps, &uap->pmh->nmaps, sizeof uap->pmh->nmaps);
    
	return error;
}

/*
 * The function called at module load/unload.
 */
static int
pmap_helper_modevt(struct module *module, int cmd, void *arg)
{
	int error = 0;

	switch (cmd) {
	case MOD_LOAD :
		printf ("pmap_helper loaded (interface version %d, syscall %d)\n", PMH_VERSION, offset);
		break;

	case MOD_UNLOAD :
		printf ("pmap_helper unloaded\n");
		break;

	default :
		error = EINVAL;
		break;
	}

	return error;
}

SYSCALL_MODULE(pmap_helper, &offset, &pmap_helper_sysent, 
    pmap_helper_modevt, NULL);
